#!/usr/bin/env python
# -*- coding: utf-8 -*-
# pylint: disable=invalid-name, too-many-arguments, broad-except
# pylint: disable=missing-docstring, line-too-long
# pylint: anomalous-backslash-in-string

import os
import array
import struct
import socket

from utils import text
from singleton import singleton

# define
DATFILE = os.path.join(os.path.dirname(__file__), "ip_addr_library.dat")
_unpack_L = lambda b: struct.unpack("<L", b)[0]
_unpack_B = lambda b: struct.unpack(">L", b)[0]


def _unpack_C(b):
    if isinstance(b, int):
        return b
    return struct.unpack("B", b)[0]


def ip2int(addr):
    return struct.unpack("!I", socket.inet_aton(addr))[0]


@singleton
class IPv4Database(object):
    """
    Database for search IPv4 address.
    The 17mon dat file format in bytes::
        -----------
        | 4 bytes |                     <- offset number
        -----------------
        | 256 * 4 bytes |               <- first ip number index
        -----------------------
        | offset - 1028 bytes |         <- ip index
        -----------------------
        |    data  storage    |
        -----------------------
    """

    def __init__(self, file_name=None):
        if file_name is None:
            file_name = DATFILE
        # store the index by the first number of ip address
        self.first_indexes = None

        # start the index ip and the correlated offset information in the db
        self.ip_num = 0
        self.ip_array = None
        self.offset_array = None
        self.length_array = None

        # store the data
        self.data = None

        self._parse(file_name)

    def _parse(self, file_name):
        buf = file(file_name, "rb").read()
        offset = _unpack_B(buf[:4])
        offset -= 1024

        self.indexes = [None] * 257
        for i in range(256):
            i_start = i * 4 + 4
            i_end = i_start + 4
            self.indexes[i] = _unpack_L(buf[i_start:i_end])
        self.indexes[256] = (offset - 1028) / 8

        self.ip_num = (offset - 1028) / 8
        self.ip_array = array.array("L", [0] * self.ip_num)
        addr_cache = dict()
        self.addr_array = [None] * self.ip_num

        for i in range(0, self.ip_num):
            pos = 1028 + 8 * i
            ip_int = _unpack_B(buf[pos:pos + 4])
            data_offset = _unpack_L(buf[pos + 4:pos + 7] + b'\0')
            data_len = _unpack_C(buf[pos + 7])

            if data_offset not in addr_cache:
                addr_cache[data_offset] = text(
                    buf[offset + data_offset:offset + data_offset + data_len])
            self.addr_array[i] = addr_cache[data_offset]
            self.ip_array[i] = ip_int

    def _lookup_ipv4(self, ip):
        nip = ip2int(ip)
        first_number = (nip >> 24) % 256
        lo, hi = self.indexes[first_number:first_number + 2]

        while lo < hi:
            mid = (lo + hi) // 2
            mid_val = self.ip_array[mid]

            if mid_val < nip:
                lo = mid + 1
            else:
                hi = mid

        if lo >= self.ip_num:
            return None
        return self.addr_array[lo]

    def find(self, ip):
        return self._lookup_ipv4(ip)


def find_ip_postion(ip):
    """
    获取 ip 地址的地理位置
    :param ip:
    :return:
    """
    return IPv4Database().find(ip)


# ------------------------
#   Demo
# ------------------------
if __name__ == '__main__':
    print(find_ip_postion('220.181.16.176'))
    print(find_ip_postion('211.161.248.29'))
    print(find_ip_postion('58.19.128.111'))
